"""hcai_nova_dynamic dataset."""
from enum import Enum
from typing import Any

import shutil
import tensorflow_datasets as tfds
import tensorflow as tf
from tensorflow_datasets.core import split_builder as split_builder_lib

from hcai_dataset_utils.bridge_tf import BridgeTensorflow
from hcai_datasets.hcai_nova_dynamic.hcai_nova_dynamic_iterable import HcaiNovaDynamicIterable
import nova_utils.db_utils.nova_types as nt

from hcai_datasets.hcai_nova_dynamic.utils.nova_string_utils import *

# TODO(hcai_audioset): Markdown description  that will appear on the catalog page.
_DESCRIPTION = """
The Nova Dynamic dataset can be used to retrieve the data and labels for a certain session or a certain part of a session of a nova dataset. 
This is part of the Nova CML Python backend (https://github.com/hcmlab/nova)
To specify which data to load use the following format: 

TODO: x
 
"""

# TODO(hcai_audioset): BibTeX citation
_CITATION = """
"""

class TFDataTypes(Enum):
  UNDEF = 0
  SHORT = tf.int16
  USHORT = tf.uint16
  INT = tf.int32
  UINT = tf.uint32
  LONG = tf.int64
  ULONG = tf.uint64
  FLOAT = tf.float32
  DOUBLE = tf.float64
  LDOUBLE = tf.float64
  BOOL = tf.bool


class HcaiNovaDynamic(HcaiNovaDynamicIterable, tfds.core.GeneratorBasedBuilder):
    """DatasetBuilder for hcai_nova_dynamic dataset."""

    VERSION = tfds.core.Version('1.0.0')
    RELEASE_NOTES = {
        '1.0.0': 'Initial release.',
    }

    def __init__(self, clear_cache=True, *args, **kwargs):
        """
        Initialize the HcaiNovaDynamic dataset builder
        Args:
          clear_cache:  when set to True the cache will be cleared else the cached dataset will be used. make sure that dataset and sample config did not change. defaults to true.
          nova_data_dir: the directory to look for data. same as the directory specified in the nova gui.
          frame_size: the framesize to look at. the matching annotation will be calculated as majority vote from all annotations that are overlapping with the timeframe.
          left_context: additional data to pass to the classifier on the left side of the frame.
          right_context: additional data to pass to the classifier on the left side of the frame.
          stride: how much a frame is moved to calculate the next sample. equals framesize by default.
          flatten_samples: if set to True samples with the same annotation scheme but from different roles will be treated as separate samples. only <scheme> is used for the keys.
          supervised_keys: if specified the dataset can be used with "as_supervised" set to True. Should be in the format <role>.<scheme>. if flatten_samples is true <role> will be ignored.
          add_rest_class: when set to True an additional restclass will be added to the end the label list
          db_config_path: path to a configfile whith the nova database config.
          db_config_dict: dictionary with the nova database config. can be used instead of db_config_path. if both are specified db_config_dict is used.
          dataset: the name of the dataset. must match the dataset name in the nova database.
          sessions: list of sessions that should be loaded. must match the session names in nova.
          annotator: the name of the annotator that labeld the session. must match annotator names in nova.
          schemes: list of the annotation schemes to fetch
          roles: list of roles for which the annotation should be loaded.
          data_streams: list datastreams for which the annotation should be loaded. must match stream names in nova.
          start: optional start time_ms. use if only a specific chunk of a session should be retreived.
          end: optional end time_ms. use if only a specifc chunk of a session should be retreived.
          **kwargs: arguments that will be passed through to the dataset builder
        """

        HcaiNovaDynamicIterable.__init__(self, *args, **kwargs)
        tfds.core.GeneratorBasedBuilder.__init__(self)

        if clear_cache:
            try:
                shutil.rmtree(self.data_dir)
            except OSError as e:
                print("Error: %s : %s" % (self.data_dir, e.strerror))

    def _info(self) -> tfds.core.DatasetInfo:
        """Returns the dataset metadata."""

        def map_label_id(lid):
            if self.flatten_samples and not lid == 'frame':
                return split_role_key(lid)[-1]
            return lid

        features_dict = {
            # TODO: Remove frame when tfds implements option to disable shuffle
            # Adding fake framenumber label for sorting
            'frame': tf.string,
            **{map_label_id(k): self._build_tfds_label_info(k, v) for k, v in self.annos.items()},
            **{map_label_id(k): self._build_tfds_data_info(k, v) for k, v in self.data_info.items()}
        }

        return tfds.core.DatasetInfo(
            builder=self,
            description=_DESCRIPTION,
            features=tfds.features.FeaturesDict(features_dict),
            supervised_keys=self.supervised_keys,
            homepage='https://github.com/hcmlab/nova',
            citation=_CITATION,
            disable_shuffling=True
        )

    def _split_generators(self, dl_manager: tfds.download.DownloadManager):
        """Returns SplitGenerators."""
        return {'dynamic_split': self._generate_examples()}

    def _generate_examples(self, **kwargs: Any) -> split_builder_lib.SplitGenerator:
        sample_counter = 1
        iter = self._yield_sample()
        while True:
            try:
                yield sample_counter, iter.__next__()
            except StopIteration:
                break
            sample_counter = sample_counter + 1

    def _build_tfds_label_info(self, label_key, scheme):

        scheme_type = self.anno_schemes[label_key]

        if scheme_type == nt.AnnoTypes.DISCRETE:
            return tfds.features.ClassLabel(names=list(scheme.labels.values()))
        elif scheme_type == nt.AnnoTypes.DISCRETE_POLYGON:
            return tfds.features.FeaturesDict(
                {
                    str(l): tfds.features.Sequence(
                        tfds.features.Tensor(shape=(2,), dtype=tf.float64)
                    )
                    for l in scheme.labels.keys()
                }
            )
        elif scheme_type == nt.AnnoTypes.FREE:
            return tfds.features.Sequence(tfds.features.Text())
        else:
            raise NotImplementedError(f"_build_tfds_label_info not implemented for {scheme_type}")

    def _build_tfds_data_info(self, data_key, data):

        if data.lazy_loading:
            return tfds.features.FeaturesDict(
                {
                    "frame_start": tf.dtypes.float32,
                    "frame_end": tf.dtypes.float32,
                    "file_path": tfds.features.Text(),
                }
            )

        scheme_type = self.data_schemes[data_key]

        if scheme_type == nt.DataTypes.VIDEO:
            return tfds.features.Sequence(
                tfds.features.Image(shape=data.sample_data_shape, dtype=tf.uint8)
            )
        elif scheme_type == nt.DataTypes.AUDIO:
            return tfds.features.Audio()
        elif scheme_type == nt.DataTypes.FEATURE:
            data_type = data.np_data_type
            if BridgeTensorflow.TYPE_MAPPING[data_type] is not None:
                data_type = BridgeTensorflow.TYPE_MAPPING[data_type]
            return tfds.features.Sequence(
                tfds.features.Tensor(shape=data.sample_data_shape, dtype=data_type)
            )
        else:
            raise NotImplementedError(f"_build_tfds_data_info not implemented for {scheme_type}")



from pathlib import Path

from nova_utils.interfaces.dataset_iterable import DatasetIterable
import numpy as np


class HcaiCkplusIterable(DatasetIterable):
    LABELS = [
        "neutral",
        "anger",
        "contempt",
        "disgust",
        "fear",
        "happy",
        "sadness",
        "suprise",
    ]

    def __init__(self, *args, dataset_dir=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.dataset_dir = Path(dataset_dir)
        self._build_index()
        self._iter = self._yield_samples(self.samples)

    def __iter__(self):
        return self._iter

    def __next__(self):
        return self._iter.__next__()

    def get_output_info(self):
        return {
            "index": {"dtype": np.str, "shape": (1,)},
            "image": {"dtype": np.str, "shape": (1,)},
            "label": {"dtype": np.uint8, "shape": (1,)},
            "rel_file_path": {"dtype": np.str, "shape": (1,)},
        }

    def _build_index(self):

        emo_anno_files = list(self.dataset_dir.glob("Emotion/**/**/*.txt"))
        samples = []

        for ef in emo_anno_files:
            with open(ef, "r") as f:
                emotion = self.LABELS[int(float(f.readline().strip()))]
                fn_img = ef.stem.replace("_emotion", ".png")
                samples.append((fn_img, emotion))

                # add neutral images
                fn_img_neut = ef.stem.rsplit("_", 2)[0] + "_00000001.png"
                samples.append((fn_img_neut, "neutral"))
        self.samples = samples

    def _yield_samples(self, files):
        """Yields examples."""

        for f, e in files:
            rel_path = (Path(*f.split("_")[:-1]) / f).resolve()
            yield {
                "index": str(f),
                "image": str(self.dataset_dir / "cohn-kanade-images" / rel_path),
                "label": self.LABELS.index(e),
                "rel_file_path": str(rel_path),
            }



from configparser import ConfigParser

import tensorflow as tf
import tensorflow_datasets as tfds
import hcai_datasets

if __name__ == "__main__":
    config = ConfigParser()
    config.read("config.ini")

    dataset, dataset_info = tfds.load(
        'hcai_ckplus',
        split='train',
        with_info=True,
        as_supervised=True,
        builder_kwargs={'dataset_dir': config["directories"]["data_dir"] + "/CK+"}
    )

    # resize images, one-hot vectors
    dataset = dataset.map(lambda x, y: (
        tf.image.resize(x, size=[224, 224]),
        tf.one_hot(y, depth=8)
    ))
    # batch
    dataset = dataset.batch(32, drop_remainder=True)

    efficientnet = tf.keras.applications.EfficientNetB0(
        include_top=True,
        weights=None,
        classes=8,
        classifier_activation="softmax"
    )
    efficientnet.compile(optimizer="adam", loss="categorical_crossentropy")
    efficientnet.fit(dataset, epochs=1)



from typing import Iterator

from torch.utils.data.dataset import T_co, IterableDataset

from hcai_dataset_utils.dataset_iterable import DatasetIterable


class BridgePyTorch(IterableDataset):
    """
    Takes a dataset iterable and pytorch transforms
    can be fed directly to a pytorch DataLoader
    """

    def __init__(self, ds_generic: DatasetIterable):
        self._ds = ds_generic
        self._transforms = {}
        self._global_transform = None

    def __iter__(self) -> Iterator[T_co]:
        return self

    def __next__(self):
        data = self._ds.__next__()
        for field in self._transforms.keys():
            data[field] = self._transforms[field](data[field])
        if self._global_transform is not None:
            data = self._global_transform(data)
        return data

    def __getitem__(self, index) -> T_co:
        return self._ds.__getitem__(index)

    def apply_global_transform(self, transform):
        self._global_transform = transform

    def apply_transform(self, field_index, transform):
        self._transforms[field_index] = transform




from pathlib import Path

from nova_utils.interfaces.dataset_iterable import DatasetIterable
import numpy as np


class HcaiFacesIterable(DatasetIterable):

    CLASSES_AGE = ["y", "o", "m"]
    CLASSES_GENDER = ["m", "f"]
    CLASSES_EMOTION = ["a", "d", "f", "h", "n", "s"]
    CLASSES_SET = ["a", "b"]

    def __init__(self, *args, dataset_dir=None, **kwargs):
        super(HcaiFacesIterable, self).__init__(*args, **kwargs)
        self.dataset_dir = Path(dataset_dir) / "bilder"
        self._generate_index()
        self._iter = self.yield_samples(self.parsed)

    def __iter__(self):
        return self._iter

    def __next__(self):
        return self._iter.__next__()

    def get_output_info(self):
        return {
            "index": {"dtype": np.str, "shape": (1,)},
            "image": {"dtype": np.str, "shape": (1,)},
            "id": {"dtype": np.uint32, "shape": (1,)},
            "age": {"dtype": np.uint8, "shape": (1,)},
            "gender": {"dtype": np.uint8, "shape": (1,)},
            "emotion": {"dtype": np.uint8, "shape": (1,)},
            "set": {"dtype": np.uint8, "shape": (1,)},
            "rel_file_path": {"dtype": np.str, "shape": (1,)},
        }

    def _generate_index(self):
        f_names = list(self.dataset_dir.glob("*.jpg"))
        parsed = [n.stem.split("_") for n in f_names]
        self.parsed = list(zip(f_names, parsed))

    def yield_samples(self, files):
        """Yields examples.
        # Labels are parsed using filenames:
        # "ID_ageGroup_gender_emotion_pictureSet.jpg"""
        for f, p in files:
            id, age, gender, emotion, set = p

            yield {
                "index": str(f),
                "image": str(f),
                "id": int(id),
                "age": self.CLASSES_AGE.index(age),
                "gender": self.CLASSES_GENDER.index(gender),
                "emotion": self.CLASSES_EMOTION.index(emotion),
                "set": self.CLASSES_SET.index(set),
                "rel_file_path": str(f.relative_to(f.parents[1])),
            }



"""hcai_ckplus dataset."""
import tensorflow as tf
import tensorflow_datasets as tfds
import pandas as pd
from tensorflow_datasets.core.splits import Split
from pathlib import Path
from hcai_dataset_utils.statistics import Statistics
from hcai_datasets.hcai_ckplus.hcai_ckplus_iterable import HcaiCkplusIterable

_DESCRIPTION = """
The Extended Cohn-Kanade (CK+) dataset contains 593 video sequences from a total of 123 different subjects, ranging from 18 to 50 years of age with a variety of genders and heritage.
"""

_CITATION = """
@inproceedings{lucey2010extended,
  title={The extended cohn-kanade dataset (ck+): A complete dataset for action unit and emotion-specified expression},
  author={Lucey, Patrick and Cohn, Jeffrey F and Kanade, Takeo and Saragih, Jason and Ambadar, Zara and Matthews, Iain},
  booktitle={2010 ieee computer society conference on computer vision and pattern recognition-workshops},
  pages={94--101},
  year={2010},
  organization={IEEE}
}
"""


class HcaiCkplus(tfds.core.GeneratorBasedBuilder, HcaiCkplusIterable, Statistics):
    """DatasetBuilder for hcai_ckplus dataset."""

    VERSION = tfds.core.Version("1.0.0")
    RELEASE_NOTES = {
        "1.0.0": "Initial release.",
    }

    def __init__(self, *args, dataset_dir, **kwargs):
        tfds.core.GeneratorBasedBuilder.__init__(self, *args, **kwargs)
        HcaiCkplusIterable.__init__(self, *args, dataset_dir=dataset_dir, **kwargs)

    def _info(self) -> tfds.core.DatasetInfo:
        """Returns the dataset metadata."""
        return tfds.core.DatasetInfo(
            builder=self,
            description=_DESCRIPTION,
            metadata=tfds.core.MetadataDict({}),
            features=tfds.features.FeaturesDict(
                {
                    "index": tf.string,
                    "image": tfds.features.Image(shape=(None, None, 3)),
                    # 0=neutral, 1=anger, 2=contempt, 3=disgust, 4=fear, 5=happy, 6=sadness, 7=surprise)
                    "label": tfds.features.ClassLabel(names=self.LABELS),
                    "rel_file_path": tf.string,
                }
            ),
            # If there's a common (input, target) tuple from the
            # features, specify them here. They'll be used if
            # `as_supervised=True` in `builder.as_dataset`.
            supervised_keys=("image", "label"),  # Set to `None` to disable
            homepage="https://dataset-homepage/",
            citation=_CITATION,
        )

    def _populate_meta_data(self, data):
        df = pd.DataFrame(data, columns=["file_name", "emotion"])
        df = df.drop(columns="file_name")
        self._populate_stats(df)

    def _split_generators(self, dl_manager: tfds.download.DownloadManager):
        """Returns SplitGenerators."""

        self._populate_meta_data(self.samples)
        return {Split.TRAIN: self._generate_examples(self.samples)}

    def _generate_examples(self, files):
        for sample in self._yield_samples(files):
            yield sample["index"], sample



import logging
import sys

import numpy as np
import os
import copy

import nova_utils.db_utils.nova_types as nt
import hcai_datasets.hcai_nova_dynamic.utils.nova_data_utils as ndu
import hcai_datasets.hcai_nova_dynamic.utils.nova_anno_utils as nau
from nova_utils.interfaces.dataset_iterable import DatasetIterable

from hcai_datasets.hcai_nova_dynamic.utils.nova_data_utils import (
    AudioData,
    VideoData,
    StreamData,
)

from hcai_datasets.hcai_nova_dynamic.nova_db_handler import NovaDBHandler
from hcai_datasets.hcai_nova_dynamic.utils.nova_string_utils import *


class HcaiNovaDynamicIterable(DatasetIterable):
    def __init__(
        self,
        *args,
        db_config_path=None,
        db_config_dict=None,
        dataset=None,
        nova_data_dir=None,
        sessions=None,
        annotator=None,
        schemes=None,
        roles=None,
        data_streams=None,
        start=None,
        end=None,
        left_context="0s",
        right_context="0s",
        frame_size=None,
        stride=None,
        add_rest_class=True,
        flatten_samples=False,
        supervised_keys=None,
        lazy_loading=False,
        fake_missing_data=True,
        **kwargs,
    ):
        """
        Initialize the HcaiNovaDynamic dataset builder
        Args:
          nova_data_dir: the directory to look for data. same as the directory specified in the nova gui.
          frame_size: the framesize to look at. the matching annotation will be calculated as majority vote from all annotations that are overlapping with the timeframe.
          left_context: additional data to pass to the classifier on the left side of the frame.
          right_context: additional data to pass to the classifier on the left side of the frame.
          stride: how much a frame is moved to calculate the next sample. equals framesize by default.
          flatten_samples: if set to True samples with the same annotation scheme but from different roles will be treated as separate samples. only <scheme> is used for the keys.
          supervised_keys: if specified the dataset can be used with "as_supervised" set to True. Should be in the format <role>.<scheme>. if flatten_samples is true <role> will be ignored.
          add_rest_class: when set to True an additional restclass will be added to the end the label list
          db_config_path: path to a configfile whith the nova database config.
          db_config_dict: dictionary with the nova database config. can be used instead of db_config_path. if both are specified db_config_dict is used.
          dataset: the name of the dataset. must match the dataset name in the nova database.
          sessions: list of sessions that should be loaded. must match the session names in nova.
          annotator: the name of the annotator that labeld the session. must match annotator names in nova.
          schemes: list of the annotation schemes to fetch
          roles: list of roles for which the annotation should be loaded.
          data_streams: list datastreams for which the annotation should be loaded. must match stream names in nova.
          start: optional start time_ms. use if only a specific chunk of a session should be retreived.
          end: optional end time_ms. use if only a specifc chunk of a session should be retreived.
          fake_missing_data: If set to true missing datastreams will provide zero data and missing annotations will be empty. If set to false a session will be skipped if an expected data stream or annotation are not found.
          **kwargs: arguments that will be passed through to the dataset builder
        """
        super().__init__(*args, **kwargs)
        self.dataset = dataset
        self.nova_data_dir = nova_data_dir
        self.sessions = sessions
        self.roles = roles
        self.schemes = schemes
        self.data_streams = data_streams
        self.annotator = annotator

        # Sliding window parameters for the main stream
        self.left_context, self.left_context_unit = ndu.parse_time_str(left_context)
        self.right_context, self.right_context_unit = ndu.parse_time_str(right_context)
        self.frame_size, self.frame_size_unit = ndu.parse_time_str(frame_size)
        if self.frame_size == 0:
            print(
                "WARNING: Frame size 0 is invalid. Returning whole session as sample."
            )
            self.frame_size = None
        self.stride, self.stride_unit = ndu.parse_time_str(stride) if stride else self.frame_size, self.frame_size_unit

        #TODO: Remove time dependent variables after refactoring
        self.left_context_ms = ndu.parse_time_string_to_ms(left_context)
        self.right_context_ms = ndu.parse_time_string_to_ms(right_context)
        self.frame_size_ms = (
            ndu.parse_time_string_to_ms(frame_size) if frame_size else None
        )

        if self.frame_size_ms == 0:
            print(
                "WARNING: Frame size 0 is invalid. Returning whole session as sample."
            )
            self.frame_size_ms = None

        self.stride_ms = (
            ndu.parse_time_string_to_ms(stride) if stride else self.frame_size_ms
        )

        self.start_ms = ndu.parse_time_string_to_ms(start)
        if not self.start_ms:
            self.start_ms = 0

        self.end_ms = ndu.parse_time_string_to_ms(end)
        if not self.end_ms:
            self.end_ms = sys.maxsize #float("inf")

        self.flatten_samples = flatten_samples
        self.add_rest_class = add_rest_class
        self.lazy_loading = lazy_loading
        self.fake_missing_data = fake_missing_data
        self.nova_db_handler = NovaDBHandler(db_config_path, db_config_dict)

        # Retrieving meta information from the database
        mongo_schemes = self.nova_db_handler.get_schemes(
            dataset=dataset, schemes=schemes
        )
        mongo_data = self.nova_db_handler.get_data_streams(
            dataset=dataset, data_streams=data_streams
        )
        self.annos, self.anno_schemes = self._populate_label_info_from_mongo_doc(
            mongo_schemes
        )
        self.data_info, self.data_schemes = self._populate_data_info_from_mongo_doc(
            mongo_data
        )

        self.session_info = {s: {} for s in self.sessions}

        # setting supervised keys
        if supervised_keys and self.flatten_samples:
            if supervised_keys[0] not in self.data_streams:
                # remove <role> of supervised keys
                _, data_stream = split_role_key(supervised_keys[0])
                if not data_stream in self.data_streams:
                    print(
                        "Warning: Cannot find supervised key '{}' in datastreams".format(
                            data_stream
                        )
                    )
                    raise Warning("Unknown data_stream")
                else:
                    supervised_keys[0] = data_stream
            if supervised_keys[1] not in self.schemes:
                # remove <role> of supervised keys
                _, scheme = split_role_key(supervised_keys[1])
                if not scheme in schemes:
                    print(
                        "Warning: Cannot find supervised key '{}' in schemes".format(
                            scheme
                        )
                    )
                    raise Warning("Unknown scheme")
                else:
                    supervised_keys[1] = scheme

        self.supervised_keys = tuple(supervised_keys) if supervised_keys else None

        self._iterable = self._yield_sample()

    def to_single_session_iterator(self, session_id=""):
        """Returns a copy of the iterator for a single session. All data objects and annotation objects will be initialized."""
        ds_iter_copy = copy.copy(self)
        if session_id not in self.sessions:
            print(
                f"WARNING: Session {session_id} not found in iterator session list. Using first session instead: {ds_iter_copy.sessions[0]}"
            )
            session_id = ds_iter_copy.sessions[0]
        elif not session_id:
            session_id = ds_iter_copy.sessions[0]

        ds_iter_copy._init_session(session_id)
        return ds_iter_copy

    def _init_session(self, session):
        """Opens all annotations and data readers"""

        if (
            not self.session_info[session]
            or not self.session_info[session]["is_active"]
        ):

            # Gather all data we need for this session
            self._load_annotation_for_session(session, time_to_ms=True)
            self._open_data_reader_for_session(session)

            session_info = self.nova_db_handler.get_session_info(self.dataset, session)[
                0
            ]

            # Depricate other sessions
            for id, s in self.session_info.items():
                s["is_active"] = False

            session_info["is_active"] = True
            self.session_info[session] = session_info

    def _populate_label_info_from_mongo_doc(self, mongo_schemes):
        """
        Setting self.annos
        Args:
          mongo_schemes:

        Returns:

        """
        annos = {}
        anno_schemes = {}

        # List of all combinations from roles and schemes that occur in the retrieved data.
        for scheme in mongo_schemes:
            for role in self.roles:
                label_id = merge_role_key(role=role, key=scheme["name"])
                scheme_type = nt.string_to_enum(nt.AnnoTypes, scheme["type"])
                scheme_name = scheme["name"]
                scheme_valid = scheme["isValid"]
                anno_schemes[label_id] = scheme_type

                if scheme_type == nt.AnnoTypes.DISCRETE:
                    labels = scheme["labels"]
                    annos[label_id] = nau.DiscreteAnnotation(
                        role=role,
                        add_rest_class=self.add_rest_class,
                        scheme=scheme_name,
                        is_valid=scheme_valid,
                        labels=labels,
                    )

                elif scheme_type == nt.AnnoTypes.CONTINUOUS:
                    min_val = scheme["min"]
                    max_val = scheme["max"]
                    sr = scheme["sr"]
                    annos[label_id] = nau.ContinuousAnnotation(
                        role=role,
                        scheme=scheme_name,
                        is_valid=scheme_valid,
                        min_val=min_val,
                        max_val=max_val,
                        sr=sr,
                    )

                elif scheme_type == nt.AnnoTypes.DISCRETE_POLYGON:
                    labels = scheme["labels"]
                    sr = scheme["sr"]
                    annos[label_id] = nau.DiscretePolygonAnnotation(
                        role=role,
                        scheme=scheme_name,
                        is_valid=scheme_valid,
                        labels=labels,
                        sr=sr,
                    )

                elif scheme_type == nt.AnnoTypes.FREE:
                    annos[label_id] = nau.FreeAnnotation(
                        role=role, scheme=scheme_name, is_valid=scheme_valid
                    )

                else:
                    raise ValueError("Invalid label type {}".format(scheme["type"]))

        return annos, anno_schemes

    def _populate_data_info_from_mongo_doc(self, mongo_data_streams):
        """
        Setting self.data
        Args:
          mongo_schemes:

        Returns:

        """
        data_info = {}
        data_schemes = {}

        for data_stream in mongo_data_streams:
            for role in self.roles:
                sample_stream_name = (
                    role + "." + data_stream["name"] + "." + data_stream["fileExt"]
                )
                sample_data_path = os.path.join(
                    self.nova_data_dir,
                    self.dataset,
                    self.sessions[0],
                    sample_stream_name,
                )
                dtype = nt.string_to_enum(nt.DataTypes, data_stream["type"])
                #window_size = (self.left_context_ms + self.frame_size_ms + self.right_context_ms) / (1 / data_stream["sr"])
                # TODO
                try:
                    if dtype == nt.DataTypes.VIDEO:
                        data = VideoData(
                            role=role,
                            name=data_stream["name"],
                            file_ext=data_stream["fileExt"],
                            sr=data_stream["sr"],
                            is_valid=data_stream["isValid"],
                            sample_data_path=sample_data_path,
                            lazy_loading=self.lazy_loading,
                        )
                    elif dtype == nt.DataTypes.AUDIO:
                        data = AudioData(
                            role=role,
                            name=data_stream["name"],
                            file_ext=data_stream["fileExt"],
                            sr=data_stream["sr"],
                            is_valid=data_stream["isValid"],
                            sample_data_path=sample_data_path,
                            lazy_loading=self.lazy_loading,
                        )
                    elif dtype == nt.DataTypes.FEATURE:
                        data = StreamData(
                            role=role,
                            name=data_stream["name"],
                            sr=data_stream["sr"],
                            data_type=dtype,
                            is_valid=data_stream["isValid"],
                            sample_data_path=sample_data_path,
                            lazy_loading=self.lazy_loading,
                        )
                    else:
                        raise ValueError(
                            "Invalid data type".format(data_stream["type"])
                        )

                except FileNotFoundError as fnf:
                    if self.fake_missing_data:
                        raise FileNotFoundError
                    else:
                        print(f"WARNING: Ignoring exception - {fnf}")
                        continue

                data_id = merge_role_key(role=role, key=data_stream["name"])
                data_info[data_id] = data
                data_schemes[data_id] = dtype

        return data_info, data_schemes

    def _load_annotation_for_session(self, session, time_to_ms=False):
        for label_id, anno in self.annos.items():
            try:
                mongo_anno = self.nova_db_handler.get_annos(
                    self.dataset, anno.scheme, session, self.annotator, anno.role
                )
            except FileNotFoundError as fnf:
                mongo_anno = []
                if self.fake_missing_data:
                    print(f"WARNING: {fnf} - Providing dummy data")
                else:
                    raise fnf
            finally:
                anno.set_annotation_from_mongo_doc(mongo_anno, time_to_ms=time_to_ms)

    def _open_data_reader_for_session(self, session):
        for data_id, data in self.data_info.items():
            try:
                data_path = os.path.join(
                    self.nova_data_dir,
                    self.dataset,
                    session,
                    data_id + "." + data.file_ext,
                )
                data.open_file_reader(data_path)
            except FileNotFoundError as fnf:
                if self.fake_missing_data:
                    print(f"WARNING: {fnf} - Providing dummy data")
                else:
                    raise fnf

    def _build_sample_dict(self, labels_for_frame, data_for_frame):
        sample_dict = {}

        garbage_detected = False
        for label_id, label_value in labels_for_frame:
            # check for nan
            if (
                    type(label_value) != list
                    and type(label_value) != str
                    and type(label_value) != np.ndarray
            ):
                try:
                    if label_value != label_value:
                        garbage_detected = True
                except:
                    breakpoint()
            sample_dict.update({label_id: label_value})

        # If at least one label is a garbage label we skip this iteration
        if garbage_detected:
            return sample_dict

        for d in data_for_frame:
            sample_dict.update(d)

        # if self.flatten_samples:
        #
        #     # grouping labels and data according to roles
        #     for role in self.roles:
        #         # filter dictionary to contain values for role
        #         sample_dict_for_role = dict(
        #             filter(lambda elem: role in elem[0], sample_dict.items())
        #         )
        #
        #         # remove role from dictionary keys
        #         sample_dict_for_role = dict(
        #             map(
        #                 lambda elem: (elem[0].replace(role + ".", ""), elem[1]),
        #                 sample_dict_for_role.items(),
        #             )
        #         )
        #
        #         sample_dict_for_role["frame"] = (
        #             str(sample_counter) + "_" + key + "_" + role
        #         )
        #         # yield key + '_' + role, sample_dict_for_role
        #         yield sample_dict_for_role
        #         sample_counter += 1
        #     c_pos_ms += _stride_ms
        #
        # else:
        return sample_dict

    def _yield_sample(self):
        """Yields examples."""

        # Needed to sort the samples later and assure that the order is the same as in nova. Necessary for CML till the tfds can be returned in order.
        sample_counter = 1

        for session in self.sessions:

            self._init_session(session)

            # Iterate through the session using frame based indices
            if self.stride_unit == ndu.StrideUnit.FRAMES:

                # Setting the main stream as a reference
                main_stream_id = self.roles[0] + '.' + self.data_streams[0]
                if main_stream_id not in self.data_info.keys():
                    print(f'WARNING: Framewise iteration has been requested but main stream "{main_stream_id}" could not be found. Skipping session.')
                    break
                print(f'Stride unit has been specified as number of frames. Using "{main_stream_id}" with a sr of {self.data_info[main_stream_id].sr} as main stream.')
                main_stream = self.data_info[main_stream_id]
                _frame_size = self.frame_size
                _stride = self.stride

                # Starting position of the first frame in seconds
                c_pos = max(self.left_context, self.start_ms)

                # If framesize is not specified we return the whole session as one junk
                if self.frame_size is None:
                    _frame_size = main_stream.n_frames
                    _stride =  main_stream.n_frames

                for i in range(0,main_stream.n_frames, _stride):
                    print('')

            else:
                session_info = self.session_info[session]
                dur = session_info["duration"]
                _frame_size_ms = self.frame_size_ms
                _stride_ms = self.stride_ms

                # If we are loading any datastreams we check if any datastream is shorter than the duration stored in the database suggests
                if self.data_info:
                    dur = min(*[v.dur for k, v in self.data_info.items()], dur)

                if not dur:
                    raise ValueError("Session {} has no duration.".format(session))

                dur_ms = int(dur * 1000)

                # If framesize is not specified we return the whole session as one junk
                if self.frame_size_ms is None:
                    _frame_size_ms = min(dur_ms, self.end_ms - self.start_ms)
                    _stride_ms = _frame_size_ms

                # Starting position of the first frame in seconds
                # c_pos_ms = self.left_context_ms + self.start_ms
                c_pos_ms = max(self.left_context_ms, self.start_ms)

                # Generate samples for this session
                while c_pos_ms + _stride_ms + self.right_context_ms <= min(
                    self.end_ms, dur_ms
                ):
                    frame_start_ms = c_pos_ms - self.left_context_ms
                    frame_end_ms = c_pos_ms + _frame_size_ms + self.right_context_ms

                    key = (
                        session
                        + "_"
                        + str(frame_start_ms / 1000)
                        + "_"
                        + str(frame_end_ms / 1000)
                    )

                    labels_for_frame = [
                        (k, v.get_label_for_frame(frame_start_ms, frame_end_ms))
                        for k, v in self.annos.items()
                    ]
                    data_for_frame = []

                    for k, v in self.data_info.items():
                        sample = v.get_sample(frame_start_ms, frame_end_ms)
                        if sample.shape[0] == 0:
                            print(f"Sample{frame_start_ms}-{frame_end_ms} is empty")
                            c_pos_ms += _stride_ms
                            continue
                        data_for_frame.append({k: sample})

                    sample_dict = self._build_sample_dict(labels_for_frame, data_for_frame)
                    if not sample_dict:
                        c_pos_ms += _stride_ms
                        sample_counter += 1
                        continue

                    yield sample_dict
                    c_pos_ms += _stride_ms
                    sample_counter += 1

            # closing file readers for this session
            for k, v in self.data_info.items():
                v.close_file_reader()

    def __iter__(self):
        return self._iterable

    def __next__(self):
        return self._iterable.__next__()

    def get_output_info(self):
        def map_label_id(lid):
            if self.flatten_samples and not lid == "frame":
                return split_role_key(lid)[-1]
            return lid

        return {
            # Adding fake framenumber label for sorting
            "frame": {"dtype": np.str, "shape": (1,)},
            **{map_label_id(k): v.get_info()[1] for k, v in self.annos.items()},
            **{map_label_id(k): v.get_info()[1] for k, v in self.data_info.items()},
        }



from hcai_dataset_utils.dataset_iterable import DatasetIterable
import numpy as np


class DummySetIterable(DatasetIterable):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._iter = self._yield_samples([
            ("a", 1, 1.0, np.array([1, 1, 1])),
            ("b", 2, 2.0, np.array([2, 2, 2])),
            ("c", 3, 3.0, np.array([3, 3, 3])),
            ("d", 4, 4.0, np.array([4, 4, 4])),
            ("e", 5, 5.0, np.array([5, 5, 5]))
        ])

    def __iter__(self):
        return self._iter

    def __next__(self):
        return self._iter.__next__()

    def get_output_info(self):
        return {
            "field_string": {"dtype": np.str, "shape": (1,)},
            "field_integer": {"dtype": np.int64, "shape": (1,)},
            "field_float": {"dtype": np.float, "shape": (1,)},
            "field_array": {"dtype": np.uint32, "shape": (3,)},
        }

    def _yield_samples(self, data):
        for sample in data:
            yield {
                "field_string": sample[0],
                "field_integer": sample[1],
                "field_float": sample[2],
                "field_array": sample[3]
            }



from configparser import ConfigParser

from hcai_dataset_utils.bridge_tf import BridgeTensorflow

import tensorflow as tf

from hcai_datasets.hcai_ckplus.hcai_ckplus_iterable import HcaiCkplusIterable

if __name__ == "__main__":

    config = ConfigParser()
    config.read("config.ini")

    iterable = HcaiCkplusIterable(
        dataset_dir=config["directories"]["data_dir"] + "/CK+",
        split="train"
    )
    dataset = BridgeTensorflow.make(iterable)

    for i, sample in enumerate(dataset):
        if i > 0:
            break
        print(sample)

    # cast to supervised tuples
    dataset = dataset.map(lambda s: (s["image"], s["label"]))
    # open files, resize images, one-hot vectors
    dataset = dataset.map(lambda x, y: (
        tf.image.resize(
            tf.image.decode_image(tf.io.read_file(x), channels=3, dtype=tf.uint8, expand_animations=False),
            size=[224, 224]
        ),
        tf.one_hot(y, depth=8)
    ))
    # batch
    dataset = dataset.batch(32, drop_remainder=True)

    for i, sample in enumerate(dataset):
        if i > 0:
            break
        print(sample[0].shape, sample[1].shape)

    efficientnet = tf.keras.applications.EfficientNetB0(
        include_top=True,
        weights=None,
        classes=8,
        classifier_activation="softmax"
    )
    efficientnet.compile(optimizer="adam", loss="categorical_crossentropy")
    efficientnet.fit(dataset, epochs=1)



from abc import ABC

from hcai_dataset_utils.dataset_iterable import DatasetIterable


class DatasetIndexed(ABC, DatasetIterable):
    """
    For Datasets which are not only iterable, but also support random access and length
    """

    def __getitem__(self, item):
        raise NotImplementedError()

    def __len__(self):
        raise NotImplementedError()



import tensorflow as tf
import numpy as np

from hcai_dataset_utils.dataset_iterable import DatasetIterable


class BridgeTensorflow:
    """
    Provides a helper function to make a tf Dataset directly from a DatasetIterable
    """

    TYPE_MAPPING = {
        np.str: tf.string,
        np.int: tf.int32,
        np.int8: tf.int8,
        np.int32: tf.int32,
        np.int64: tf.int64,
        np.uint: tf.uint32,
        np.uint8: tf.uint8,
        np.uint32: tf.uint32,
        np.uint64: tf.uint64,
        np.float: tf.float32,
        np.float32: tf.float32,
        np.float64: tf.float32,
    }

    @staticmethod
    def make(ds: DatasetIterable):
        iter = ds.__iter__

        info = ds.get_output_info()
        output_types = {
            **{k: BridgeTensorflow.TYPE_MAPPING[info[k]["dtype"]] for k in info.keys()}
        }

        return tf.data.Dataset.from_generator(iter, output_types=output_types)



from pathlib import Path
from hcai_dataset_utils.bridge_tf import BridgeTensorflow
import tensorflow as tf
from hcai_datasets.hcai_nova_dynamic.hcai_nova_dynamic_iterable import HcaiNovaDynamicIterable

if __name__ == "__main__":

    iterable = HcaiNovaDynamicIterable(
        db_config_path="../../local/nova_db.cfg",
        db_config_dict=None,
        # Dataset Config
        dataset="affect-net",
        nova_data_dir=Path("Z:\\nova\\data"),
        sessions=[f"{i}_man_eval" for i in range(8)],
        roles=["session"],
        schemes=["emotion_categorical"],  # [""],
        # schemes=["neuthappysad"],
        annotator="gold",
        # annotator="system",
        data_streams=["video"],
        # Sample Config
        frame_size=0.04,
        left_context=0,
        right_context=0,
        # start = "0s",
        # end = "300s",
        flatten_samples=False,
        # Additional Config
        lazy_loading=False
    )
    dataset = BridgeTensorflow.make(iterable)

    for i, sample in enumerate(dataset):
        if i > 0:
            break
        print(sample["session.video"].shape)
        print(sample["session.emotion_categorical"].shape)

    # cast to supervised tuples
    dataset = dataset.map(lambda s: (s["session.video"], s["session.emotion_categorical"]))
    # resize images, one-hot vectors
    dataset = dataset.map(lambda x, y: (
        tf.image.resize(tf.ensure_shape(tf.squeeze(x), [300, 300, 3]), [224, 224]),
        tf.one_hot(y, depth=11)
    ))
    # batch
    dataset = dataset.batch(32, drop_remainder=True)

    for i, sample in enumerate(dataset):
        if i > 0:
            break
        print(sample[0].shape, sample[1].shape)

    efficientnet = tf.keras.applications.EfficientNetB0(
        include_top=True,
        weights=None,
        classes=11,
        classifier_activation="softmax"
    )
    efficientnet.compile(optimizer="adam", loss="categorical_crossentropy")
    efficientnet.fit(dataset, epochs=1)



from unittest import TestCase

import torch
from torch.utils.data import DataLoader

from hcai_dataset_utils.bridge_pytorch import BridgePyTorch
from hcai_datasets.tests.dummy_set import DummySetIterable


class TestBridgePyTorch(TestCase):

    def test_bridge(self):
        dataset = DummySetIterable()

        loader = DataLoader(BridgePyTorch(dataset))

        for sample in loader:
            self.assertEqual(sample["field_float"].dtype, torch.float64)
            self.assertEqual(sample["field_integer"].dtype, torch.int64)
            self.assertEqual(sample["field_array"].shape[1], 3)



from configparser import ConfigParser

from torch.utils.data import DataLoader

from hcai_dataset_utils.bridge_pytorch import BridgePyTorch
from hcai_datasets.hcai_affectnet.hcai_affectnet_iterable import HcaiAffectnetIterable

if __name__ == "__main__":

    config = ConfigParser()
    config.read("config.ini")

    iterable = HcaiAffectnetIterable(
        dataset_dir=config["directories"]["data_dir"] + "/AffectNet",
        split="test"
    )
    dataloader = DataLoader(BridgePyTorch(iterable))

    for i, sample in enumerate(dataloader):
        if i > 0:
            break
        print(sample)




import ast
from pathlib import Path
from importlib.util import find_spec


def get_imports(file_path):
    """Returns a list of imported packages for the respective module"""
    with open(file_path, "r") as f:
        node = ast.parse(f.read())
    imports = [n.names[0].name for n in node.body if isinstance(n, ast.Import)]
    return imports


def validate_installation(file_path: Path):
    """Checks if every dependency in the list is installed"""
    dependencies = get_imports(file_path)
    module_id = file_path.name[:-3]

    for dependency in dependencies:
        if not find_spec(dependency):
            print(f"Package {dependency} not installed. Import of {module_id} skipped")
            return False
    return True



"""A setuptools based setup module.
See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
import datetime
import sys
import pkg_resources
import pathlib
from version import __version__

# Parse arguments
if "--nightly" in sys.argv:
    nightly = True
    sys.argv.remove("--nightly")
else:
    nightly = False

# Settings
project_name = "hcai-datasets"
here = pathlib.Path(__file__).parent.resolve()
long_description = (here / "README.md").read_text(encoding="utf-8")


# Adjustment for nightly build
if nightly:
    project_name += "-nightly"
    datestring = datetime.datetime.now().strftime("%Y%m%d%H%M")
    curr_version = pkg_resources.parse_version(__version__)
    __version__ = f"{curr_version.base_version}.dev{datestring}"

# Setup
setup(
    name=project_name,
    version=__version__,
    description="!Alpha Version! - This repository contains the backend server for the nova annotation ui (https://github.com/hcmlab/nova)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hcmlab/nova-server",
    author="Dominik Schiller",
    author_email="dominik.schiller@uni-a.de",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
    ],
    packages=find_packages(),
    include_package_data=True,
    python_requires=">=3.6, <4",
    install_requires=[
        "pymongo==3.12",
        "decord;platform_machine=='AMD64'",
        "numba",
        "pandas==1.4"
    ],
    extras_require={
        "torch": ["torch"],
        "tf" : [  "tensorflow==2.7", "tensorflow-datasets==4.4.0"]
    }
)
